#!/usr/bin/env python
# ===============================================================================
# Copyright (c) 2016 by Frank Fischer
# ===============================================================================
from __future__ import unicode_literals, print_function, with_statement

try:
    from subprocess import STDOUT, check_output, CalledProcessError
except ImportError:
    # python 2.6 doesn't include check_output. Monkey patch it in!
    import subprocess

    STDOUT = subprocess.STDOUT


    def check_output(*popenargs, **kwargs):
        if "stdout" in kwargs:  # pragma: no cover
            raise ValueError("stdout argument not allowed, it will be overridden.")
        process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs, **kwargs)
        output, _ = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise subprocess.CalledProcessError(retcode, cmd, output=output)
        return output


    subprocess.check_output = check_output


    class CalledProcessError(Exception):
        # overwrite CalledProcessError due to `output` keyword not being available (in 2.6)

        def __init__(self, returncode, cmd, output=None):
            self.returncode = returncode
            self.cmd = cmd
            self.output = output

        def __str__(self):
            return "Command '%s' returned non-zero exit status %d" % (self.cmd, self.returncode)


    subprocess.CalledProcessError = CalledProcessError

from shutil import rmtree
import argparse
import json
import os
import errno
import shlex
import subprocess
import syslog
from distutils.spawn import find_executable


class GitConfigPull(object):
    file_list_key = "config"
    script_key = "scripts"
    script_name = "name"
    script_destination = "destination"

    def __init__(self, **kwargs):
        """Pull HTCondor configuration from a remote repository during condor_reconfig."""

        def clear_path(path_):
            """Convert path to absolute path & normalize it"""
            return os.path.abspath(os.path.join(os.path.normpath(path_), ""))

        if find_executable("git") is None:
            syslog.syslog(syslog.LOG_ERR, "'git' not found in %s" % os.environ['PATH'])
            raise RuntimeError("Could not find 'git' on system (Paths: %s)." % os.environ['PATH'])

        self._git_remote = kwargs.get("remote")
        self._git_branch = kwargs.get("branch")
        self._json_filename = clear_path(kwargs.get("input"))
        self._json_key = kwargs.get("site")
        self._git_folder = clear_path(kwargs.get("git"))
        self._out = clear_path(kwargs.get("out"))
        self._depth = kwargs.get("depth")

    def main(self):
        """Main execution method"""
        self._git_sync()

        self._create_links()

        git_head = self.get_git_head(self._git_folder)
        if git_head is not None:
            syslog.syslog("Success, HEAD is %s" % git_head)
            print("GitHead=\"%s\"" % git_head.rstrip())

    def _git_sync(self):
        """Clone repository // Clean, reset & pull."""
        try:
            if os.path.exists(os.path.join(self._git_folder, ".git")):
                self.git_pull(repository_folder=self._git_folder)
            else:
                # Check parent folder(s) for existence
                if not os.path.exists(os.path.dirname(self._git_folder)):
                    os.makedirs(os.path.dirname(self._git_folder))
                else:
                    rmtree(self._git_folder)

                self.git_clone(remote=self._git_remote, branch=self._git_branch, output_folder=self._git_folder,
                               depth=self._depth)

        except subprocess.CalledProcessError as err:
            syslog.syslog(syslog.LOG_ERR, "Command %s failed with RC %d. Output: %s" %
                          (err.cmd, err.returncode, err.output))
        except AssertionError:
            # In case of problems, remove the repository and do a re-clone.
            rmtree(self._git_folder)
            self._git_sync()

    def _create_links(self):
        """Read JSON file and create corresponding links."""
        try:
            ###
            # Parse JSON file
            ###
            config_files, script_files = self._read_json(json_file=self._json_filename, json_key=self._json_key)
            ###
            # Config files
            # Linked from "repository"-folder to "out"-folder
            ###
            for entry in config_files:
                out_file = os.path.join(self._git_folder, entry)
                if not os.path.exists(out_file):
                    syslog.syslog(syslog.LOG_WARNING, "File %s does not exist. Skipping link." % out_file)
                    continue

                syslog.syslog("Linking %s to %s" % (entry, self._out))
                try:
                    # Link absolute paths
                    symlink_force(os.path.join(self._git_folder, entry), os.path.join(self._out, entry))
                except OSError as error:
                    syslog.syslog(syslog.LOG_ERR, str(error))
            ###
            # Scripts
            # Linked to (explicit) script destination per file
            ###
            for line in script_files:
                out_file = os.path.join(self._git_folder, "scripts", line[self.script_name])
                if not os.path.exists(out_file):
                    syslog.syslog(syslog.LOG_WARNING, "File %s does not exist. Skipping link." % out_file)
                    continue
                syslog.syslog("Linking scripts/%s to %s" % (line[self.script_name], line[self.script_destination]))
                try:
                    symlink_force(out_file, os.path.join(line[self.script_destination], line[self.script_name]))
                except OSError as error:
                    syslog.syslog(syslog.LOG_ERR, str(error))
        except (IOError, KeyError) as err:
            syslog.syslog(syslog.LOG_ERR, str(err))
            # Raise file error RC (errno.ENOENT)
            exit(2)

    @classmethod
    def _read_json(cls, json_file, json_key):
        # type: (string, string) -> Tuple[list, list[dict]]
        """Open JSON file and return a single key's values."""
        try:
            with open(json_file, mode="r") as json_file:
                result = json.load(json_file)
                configs = result[json_key][cls.file_list_key]
                scripts = result[json_key].get(cls.script_key, list())
                return configs, scripts
        except (IOError, KeyError) as err:
            if not os.path.exists(json_file.name):
                raise IOError("Could not read file '%s'." % json_file)
            else:
                raise KeyError("Error when parsing '%s': %s." % (json_file, err))

    @staticmethod
    def git_clone(remote, branch, output_folder, depth=0):
        # type: (string, string, string) -> None
        """Clone repository into output_folder.
         Redirects push remote to /dev/null.
         """
        syslog.syslog("Cloning repository %s (%s) into %s" % (remote, branch, output_folder))
        if depth:
            option = "--depth %d" % depth
        else:
            option = ""
        subprocess.check_call(shlex.split("git clone -q %s %s --branch %s %s" %
                                          (option, remote, branch,
                                           output_folder)))
        os.chdir(output_folder)
        subprocess.check_call(shlex.split("git remote set-url --push origin /dev/null"))

    @staticmethod
    def git_pull(repository_folder):
        # type: (string) -> None
        """Perform 'git pull' (including hard reset on folder)."""
        os.chdir(repository_folder)
        try:
            assert subprocess.check_output(shlex.split("git remote")) != "", "No (remote) repository defined."
        except subprocess.CalledProcessError as err:
            if err.returncode != 0:
                raise AssertionError("No (remote) repository defined.")
        syslog.syslog("Clean & reset repository + pull")
        subprocess.check_call(shlex.split("git clean -q -f -d"))
        subprocess.check_call(shlex.split("git reset -q --hard"))
        subprocess.check_call(shlex.split("git pull -q -f"))

    @staticmethod
    def get_git_head(repository_folder):
        # type: (string) -> string
        """Get Git HEAD (hash)."""
        try:
            os.chdir(repository_folder)
            return subprocess.check_output(shlex.split("git rev-parse HEAD"))
        except subprocess.CalledProcessError as err:
            syslog.syslog(syslog.LOG_ERR, "Failure during sync: %s" % err.output)


def symlink_force(source, link_name):
    # type: (Union[str, unicode], Union[str, unicode]) -> None
    """Wrapper to os.symlink, emulating ln -fs.

    Included convenience functions :
     - create intermediate dirs to target
     - link name can be a directory
     - No action, if link already exists
    """
    try:
        os.makedirs(os.path.dirname(link_name))
    except OSError:
        pass

    if os.path.isdir(link_name):
        link_destination = os.path.join(link_name, os.path.basename(source))
    else:
        link_destination = link_name

    if os.path.exists(link_destination):
        if os.path.samefile(source, link_destination):
            return

    try:
        os.symlink(source, link_destination)
    except OSError as e:
        if e.errno == errno.EEXIST:
            os.remove(link_destination)
            os.symlink(source, link_destination)
        else:
            raise e


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Pull HTCondor configuration from a remote repository during "
                                                 "condor_reconfig and output CONFIG_HASH. This script tries to "
                                                 "load the detailed configuration composition from a JSON file.")
    parser.add_argument("-r", "--remote", default="https://gitlab.etp.kit.edu/EKPCondor/HTConfig.git", type=str,
                        help="Remote Git repository where all files are stored. (default: %(default)s)")
    parser.add_argument("-b", "--branch", default="master", type=str,
                        help="Refspec/branch to check out from repository. (default: %(default)s)")
    parser.add_argument("-d", "--depth", default=0, type=int,
                        help="Check out repository with a limited amount of history (0 = all, 1 = no history, "
                             "2 = a single previous entry, ...). (default: %(default)s)")
    parser.add_argument("-i", "--input", default="/etc/condor/config_sites.json", type=str,
                        help="JSON file which contains all site combinations. (default: %(default)s)")
    parser.add_argument("-g", "--git", default="/etc/condor/git/", type=str,
                        help="Local directory of the Git repository. "
                             "Will be cloned into or updated (default: %(default)s)")
    parser.add_argument("-o", "--out", default="/etc/condor/config.d/", type=str,
                        help="Local directory with configuration files. (default: %(default)s)")
    parser.add_argument("site", type=str, help="Site name (corresponding to JSON file keys).")
    try:
        GitConfigPull(**vars(parser.parse_args())).main()
    except RuntimeError as e:
        # Raise missing package error (errno.ENOPKG)
        exit(65)
